import re
from typing import Any, Dict, List, Union

from lark import Token
from lark.visitors import Transformer


class BaseTransformer(Transformer):
    def NUM_ID(self, num):
        return num.update(value=int(num))

    def TOKEN(self, tok):
        return tok.update(value=tok.strip())

    def protein_id(self, children) -> Token:
        children = [x.value for x in children]
        return Token("protein_id", children)

    def ref_id(self, children) -> Token:
        children = [x.value for x in children]
        return Token("ref_id", children)

    @staticmethod
    def _fix_string(s: str):
        s = re.sub(r"([\[(]) ", r"\1", s)
        s = re.sub(r" ([\])])", r"\1", s)
        s = s.replace("\t", " ")
        return s


class GenericTreeTransformer(BaseTransformer):
    """Transform extracted values from bottom-up.

    Formats the tree into a dictionary, as described in :meth:`.entry`.
    """

    def description(self, children) -> Token:
        res = " ".join([x.value.strip() for x in children])
        res = self._fix_string(res)
        return Token("description", res)

    def content(self, children) -> Dict[str, Union[str, List[int]]]:
        res = {x.type: x.value for x in children}
        return res

    def commentary(self, children) -> List[Dict[str, Union[str, List[int]]]]:
        return Token("commentary", children)

    def entry(self, children: List[Token]) -> Dict[str, Any]:
        res = {x.type: x.value for x in children}
        return res


class ReactionTreeTransformer(GenericTreeTransformer):
    """
    Commentary in ``(...)`` are on substrates, and in ``|...|`` on products.
    """

    def reaction(self, children) -> Token:
        """Parse the reaction in the description node.

        There should be three parts:
            - The left-hand-side of the reaction, which is a list of chemical
                names, separated by ``<space>+<space>``.
            - The separator ``<space>=<space>``.
            - The right-hand-side of the reaction, which is a list of chemical
                names, separated by ``<space>+<space>``.
        """
        reaction = " ".join([x.value.strip() for x in children])
        reaction = self._fix_string(reaction)
        if " = " not in reaction:
            return Token("reaction_description", reaction)
        lhs, rhs = reaction.split(" = ")
        lhs = lhs.split(" + ")
        rhs = rhs.split(" + ")
        res = {"lhs": lhs, "rhs": rhs}
        return Token("reaction", res)

    def commentary(self, children) -> Token:
        return Token("commentary_substrate", children)

    def more_commentary(self, children) -> Token:
        return Token("commentary_product", children)

    def reversibility(self, children) -> Token:
        """
        ``r`` for reversible, ``ir`` for irreversible, ``?`` for unknown.
        """
        x = children[0].value
        x = x.replace("{", "")
        x = x.replace("}", "")
        if not x:
            x = "?"
        return Token("reversibility", x)

    def entry(self, children: List[Token]) -> Dict[str, Any]:
        res = {x.type: x.value for x in children}
        if "reversibility" in res:
            res["reaction"]["reversibility"] = res["reversibility"]
            del res["reversibility"]
        return res


class RefTreeTransformer(BaseTransformer):
    def ref_id(self, children) -> Token:
        return Token("ref_id", children[0].value)

    def citation(self, children) -> Token:
        res = " ".join([x.value.strip() for x in children])
        res = self._fix_string(res)
        return Token("citation", res)

    def pubmed(self, children) -> Token:
        res = children[0].value
        res = re.sub(r"\{Pubmed:(\d*)\}", r"\1", res)
        return Token("pubmed", res)

    def paper_stat(self, children) -> Token:
        res = children[0].split(",")
        return Token("paper_stat", res)

    def entry(self, children: List[Token]) -> Dict[str, Any]:
        return {x.type: x.value for x in children}


class SpecificInfoTreeTransformer(GenericTreeTransformer):
    def substrate(self, children) -> Token:
        res = children[0].value.strip()
        return Token("substrate", res)


class CommentaryOnlyTreeTransformer(BaseTransformer):
    def description(self, children) -> Token:
        res = children[0]
        res = self._fix_string(res)
        res = res[1:-1]  # remove the brackets
        return Token("description", res)

    def entry(self, children: List[Token]) -> Dict[str, Any]:
        res = {x.type: x.value for x in children}
        return res



import logging
from typing import Any, Dict, List, Tuple, Union

from metabolike.db.base import BaseDB

logger = logging.getLogger(__name__)

NODE_LABELS = [
    "Pathway",
    "Compartment",
    "Reaction",
    "Compound",
    "GeneProduct",
    "RDF",
    "Complex",
    "EntitySet",
    "Citation",
    "Taxa",
]

COMMON_COMPOUNDS = ["ATP", "ADP", "H+", "NADH", "NAD+", "H2O", "phosphate"]


class MetaDB(BaseDB):
    def use_database(self, db_name: str):
        """Set the name of the database to use."""
        self.db_name = db_name
        self.start_session(database=self.db_name)

    def setup_graph_db(self, create_db: bool = True, **kwargs):
        """
        Create Neo4j database and set proper constraints. ``Reaction`` nodes are
        central in the schema.

        Args:
            session: :class:`neo4j.Session` object.
            create_db: If False, does not create the database. This is useful
                for running on neo4j AuraDB when database creation is not allowed.
            **kwargs: Keyword arguments to pass to :meth:`.create`.
        """
        # Create database
        if create_db:
            self.create(self.db_name, **kwargs)

        # Set constraints
        logger.debug("Creating constraint for RDF nodes")
        r = self.write(
            """CREATE CONSTRAINT IF NOT EXISTS
                 ON (r:RDF) ASSERT r.uri IS UNIQUE;""",
        ).data()
        if r:
            logger.warning(f"Could not create constraint for RDF nodes: {r}")

        # Constraints automatically create indexes, so we don't need to
        # create them manually.
        for label in NODE_LABELS:
            logger.debug(f"Creating constraint for node: {label}")
            r = self.write(
                f"""CREATE CONSTRAINT IF NOT EXISTS
                    ON (n:{label}) ASSERT n.mcId IS UNIQUE;""",
            ).data()
            if r:
                logger.warning(f"Could not create constraint for node {label}: {r}")

    def create_node(self, node_label: str, mcid: str, props: Dict[str, str]):
        """Create a node with the given label and properties.

        See :meth:`.Metacyc.sbml_to_graph` for details.

        Args:
            node_label: Label of the node.
            mcid: ``mcId`` of the node.
            props: Properties of the node.
        """
        logger.debug(f"Creating {node_label} node: {mcid}")
        self.write(
            f"""
            MERGE (n:{node_label} {{mcId: $mcId}})
            ON CREATE SET n += $props;
            """,
            mcId=mcid,
            props=props,
        )

    def create_compound_node(self, compartment: str, mcid: str, props: Dict[str, str]):
        """See :meth:`.Metacyc.sbml_to_graph` for details."""
        logger.debug(f"Creating Compound node: {mcid}")
        self.write(
            """
            MATCH (cpt:Compartment {mcId: $compartment})
            MERGE (c:Compound {mcId: $mcId})-[:hasCompartment]->(cpt)
            ON CREATE SET c += $props;
            """,
            compartment=compartment,
            mcId=mcid,
            props=props,
        )

    def link_node_to_rdf(
        self,
        node_label: str,
        mcid: str,
        bio_qual: str,
        props: Dict[str, Union[str, List[str]]],
    ):
        """See :meth:`.Metacyc._add_sbml_rdf_node` for details."""
        logger.debug(f"{node_label} node {mcid} {bio_qual} RDF")
        self.write(
            f"""
                MATCH (c:{node_label} {{mcId: $mcId}})
                MERGE (n:RDF)<-[:{bio_qual}]-(c)
                ON CREATE SET n += $props;
                """,
            mcId=mcid,
            props=props,
        )

    def link_reaction_to_compound(
        self,
        reaction_id: str,
        compound_id: str,
        compound_type: str,
        props: Dict[str, str],
    ):
        logger.debug(f"Adding {compound_type} {compound_id} to Reaction {reaction_id}")
        self.write(
            f"""
            MATCH (r:Reaction {{mcId: $reaction}}),
                  (c:Compound {{mcId: $compound}})
            MERGE (r)-[l:has{compound_type}]->(c)
            ON CREATE SET l += $props;
            """,
            reaction=reaction_id,
            compound=compound_id,
            props=props,
        )

    def link_reaction_to_compartment(self, reaction_id: str, compartment_name: str):
        self.write(
            """
            MATCH (r:Reaction {displayName: $reaction}),
                  (c:Compartment {displayName: $compartment})
            MERGE (r)-[:hasCompartment]->(c);
            """,
            reaction=reaction_id,
            compartment=compartment_name,
        )

    def link_node_to_gene(
        self,
        node_label: str,
        node_id: str,
        group_id: str,
        group_label: str,
        edge_type: str,
    ):
        """
        See :meth:`.Metacyc._add_sbml_gene_product_association_node` for details.

        Link a node to a ``GeneProduct``, ``Complex``, or ``EntitySet`` node.

        Args:
            node_label: Label of the node.
            node_id: ``mcId`` of the node.
            group_id: ``mcId`` of the ``GeneProduct``, ``Complex`` or ``EntitySet``.
            group_label: ``Complex`` or ``EntitySet``.
            edge_type: Type of the edge.
        """
        if group_label not in ["GeneProduct", "Complex", "EntitySet"]:
            raise ValueError(f"Invalid group_type: {group_label}")
        logger.debug(
            f"{node_label} node {node_id} {edge_type} {group_label} {group_id}"
        )
        self.write(
            f"""
            MATCH (n:{node_label} {{mcId: $node_id}})
            MERGE (g:{group_label} {{mcId: $group_id}})
            MERGE (g)<-[l:{edge_type}]-(n)
            """,
            node_id=node_id,
            group_id=group_id,
        )

    def get_all_nodes(self, label: str, prop: str) -> List[str]:
        """Fetch an property of nodes with a certain label."""
        if label not in NODE_LABELS:
            raise ValueError(f"Invalid label: {label}")
        # TODO: check for valid properties
        res = self.read(
            f"""
            MATCH (n:{label})
            RETURN DISTINCT n.{prop} AS prop;
            """
        )
        return [n["prop"] for n in res]

    def get_all_compounds(self) -> List[Tuple[str, str]]:
        """
        Fetch all ``Compound`` nodes. All Compound node with RDF have BioCyc IDs.
        """
        res = self.read(
            """
            MATCH (c:Compound)-[:is]->(r:RDF)
            RETURN DISTINCT c.displayName, r.biocyc;
            """
        )  # TODO: 38 POLYMER nodes don't have BioCyc IDs
        return [(cpd["c.displayName"], cpd["r.biocyc"]) for cpd in res]

    def link_node_to_citation(
        self, node_label: str, node_display_name: str, citation_id: str
    ):
        """Link a node to a ``Citation`` node.

        Args:
            node_label: Type of the node (Reaction, Pathway, or Compound).
            node_display_name: ``displayName`` of the node.
            citation_id: ``mcId`` of the ``Citation`` node.
        """
        logger.debug(
            f"{node_label} node {node_display_name} has citation {citation_id}"
        )
        self.write(
            f"""
            MATCH (n:{node_label} {{displayName: $dn}})
            MERGE (c:Citation {{mcId: $citation}})
            MERGE (n)-[:hasCitation]->(c)
            """,
            dn=node_display_name,
            citation=citation_id,
        )

    def add_props_to_node(
        self,
        node_label: str,
        node_prop_key: str,
        node_prop_val: str,
        props: Dict[str, Any],
    ):
        """Add properties to a node.

        Args:
            node_label: Label of the node.
            node_prop_key: Property key of the node used to locate the node.
            node_prop_val: Property value of the node used to filter ``node_prop_key``.
            props: Properties to add to the node.
        """
        if node_label not in NODE_LABELS:
            raise ValueError(f"Invalid label: {node_label}")
        # TODO: check for valid properties
        self.write(
            f"""
            MATCH (n:{node_label} {{{node_prop_key}: $val}})
            SET n += $props;
            """,
            val=node_prop_val,
            props=props,
        )

    def add_props_to_compound(
        self, compound_id: str, c_props: Dict[str, Any], rdf_props: Dict[str, Any]
    ):
        self.write(
            """
            MATCH (c:Compound {displayName: $cpd_id})-[:is]->(r:RDF)
            SET c += $c_props, r += $rdf_props;
            """,
            cpd_id=compound_id,
            c_props=c_props,
            rdf_props=rdf_props,
        )

    def link_reaction_to_pathway(self, reaction_id: str, pathway_id: str):
        logger.debug(f"Reaction {reaction_id} is in Pathway {pathway_id}")
        self.write(
            """
            MATCH (r:Reaction {displayName: $reaction})
            MERGE (pw:Pathway {mcId: $pathway})
                ON CREATE SET pw.displayName = $pathway
            MERGE (pw)-[:hasReaction]->(r)
            """,
            reaction=reaction_id,
            pathway=pathway_id,
        )

    def link_pathway_to_superpathway(self, pathway_id: str, superpathway_id: str):
        """Link a ``Pathway`` to its super-pathway."""
        logger.debug(f"Pathway {pathway_id} has SuperPathway {superpathway_id}")
        self.write(
            """
            MATCH (pw:Pathway {displayName: $pw_id})
            MERGE (spw:Pathway {mcId: $super_pw})
                ON CREATE SET spw.displayName = $super_pw
            MERGE (spw)-[:hasSubPathway]->(pw)
            """,
            pw_id=pathway_id,
            super_pw=superpathway_id,
        )

    def link_pathway_to_pathway(
        self, pw: str, pws: List[str], direction: str, cpd: str
    ):
        """
        Connect a ``Pathway`` to a list of other ``Pathway``s. The ``Compound``
        that is involved in both sides of the connection is specified in the
        relationship.
        """
        if direction == "INCOMING":
            rel_type = "-[l:hasRelatedPathway]->"
        else:
            rel_type = "<-[l:hasRelatedPathway]-"
        self.write(
            f"""
            MATCH (pw1:Pathway {{displayName: $pw_id}})
            MATCH (pw2s:Pathway) WHERE pw2s.displayName IN $pws
            UNWIND pw2s AS pw2
            MERGE (pw1){rel_type}(pw2)
                ON CREATE SET l.hasRelatedCompound = $cpd
            """,
            pw_id=pw,
            pws=pws,
            cpd=cpd,
        )

    def link_pathway_to_taxa(self, pathway_id: str, tax_id: str, relationship: str):
        """Link a ``Pathway`` to a ``Taxa`` node."""
        if not relationship in {"hasRelatedSpecies", "hasExpectedTaxonRange"}:
            raise ValueError(f"Invalid relationship: {relationship}")
        logger.debug(f"Pathway {pathway_id} {relationship} {tax_id}")
        self.write(
            f"""
            MATCH (pw:Pathway {{displayName: $pw}})
            MERGE (s:Taxa {{mcId: $taxa_id}})
            MERGE (pw)-[:{relationship}]->(s)
            """,
            pw=pathway_id,
            taxa_id=tax_id,
        )

    def link_pathway_to_rate_limiting_step(self, pathway_id: str, reaction_id: str):
        logger.debug(f"Pathway {pathway_id} has rate limiting step {reaction_id}")
        self.write(
            """
            MATCH (pw:Pathway {displayName: $pw}),
                  (r:Reaction {canonicalId: $rxn})
            MERGE (pw)-[l:hasReaction]->(r)
                ON MATCH SET l.isRateLimitingStep = true
            """,
            pw=pathway_id,
            rxn=reaction_id,
        )

    def link_pathway_to_primary_compound(
        self, pathway_id: str, compound_id: str, relationship: str
    ):
        logger.debug(f"Pathway {pathway_id} has primary {relationship} {compound_id}")
        self.write(
            f"""
            MATCH (cpd:Compound)-[:is]->(:RDF {{Biocyc: $compound_id}}),
                  (pw:Pathway {{displayName: $pw}})-[:hasReaction]->(:Reaction)-[:hasLeft|hasRight]->(cpd)
            MERGE (pw)-[:hasPrimary{relationship}]->(cpd)
            """,
            pw=pathway_id,
            compound_id=compound_id,
        )

    def link_reaction_to_next_step(self, r1: str, r2: List[str], pathway_id: str):
        """Link a reaction to its next step in a pathway."""
        logger.debug(f"Reaction {r1} has next steps {r2} in {pathway_id}")
        self.write(
            """
            MATCH (r1:Reaction {canonicalId: $r1})
            MATCH (r2:Reaction) WHERE r2.canonicalId IN $r2
            UNWIND r2 AS pred
            MERGE (pred)-[l:isPrecedingEvent]->(r1)
                ON CREATE SET l.hasRelatedPathway = $pw
            """,
            r1=r1,
            r2=r2,
            pw=pathway_id,
        )

    def link_reaction_to_primary_compound(
        self,
        reaction_id: str,
        compound_ids: List[str],
        pathway_id: str,
        side: str,
    ):
        logger.debug(f"Reaction {reaction_id} has primary {side}s {compound_ids}")
        self.write(
            f"""
            MATCH (r:Reaction {{canonicalId: $rxn_id}}),
                  (cpds:Compound)-[:is]->(rdf:RDF)
            WHERE rdf.biocyc IN $compound_ids
            UNWIND cpds AS cpd
            MATCH (r)-[l:hasLeft|hasRight]->(cpd)
            SET l.isPrimary{side}InPathway = CASE
              WHEN l.isPrimary{side}InPathway IS NULL THEN [$pw]
              WHEN $pw IN l.isPrimary{side}InPathway THEN l.isPrimary{side}InPathway
              ELSE l.isPrimary{side}InPathway + [$pw]
              END;
            """,
            rxn_id=reaction_id,
            compound_ids=compound_ids,
            pw=pathway_id,
        )

    def merge_nodes(
        self, n1_label: str, n2_label: str, n1_attr: str, n2_attr: str, attr_val: str
    ):
        """
        Merge two nodes with the same attribute value. Note that properties is
        hard-coded to be "override", which means all node attributes of ``n2``
        will be used to override the attributes of ``n1``.

        Args:
            n1_label: The label of the first node.
            n2_label: The label of the second node.
            n1_attr: The attribute of the first node for filtering.
            n2_attr: The attribute of the second node for filtering.
            attr_val: The value of the attributes for filtering.
        """
        _ = self.write(
            f"""
            MATCH (n1:{n1_label} {{{n1_attr}: $val}}),
                  (n2:{n2_label} {{{n2_attr}: $val}})
            CALL apoc.refactor.mergeNodes([n1, n2], {{properties: 'override'}})
            YIELD node
            RETURN node;
            """,
            val=attr_val,
        )

    def fix_reaction_direction(self):
        """
        Seems like ``listOfReactants`` in the SBML file are always the true
        reactants in irreversible reactions, no matter if they are under
        ``LEFT`` or ``RIGHT`` in the dat file. Since the reaction nodes were all
        populated using the SBML file, we can say all reactions are either
        reversible or go from left to right.
        """
        self.write(
            """
            MATCH (r:Reaction {reactionDirection: 'right_to_left'})
            SET r.reactionDirection = 'left_to_right'
            """
        )
        self.write(
            """
            MATCH (r:Reaction {reactionDirection: 'physiol_right_to_left'})
            SET r.reactionDirection = 'physiol_left_to_right'
            """
        )

    def delete_bad_reaction_nodes(self):
        """
        Delete reaction nodes with non-canonical mcIds and all their relationships.
        """
        self.write(
            """
            MATCH (r:Reaction) WHERE r.displayName <> r.canonicalId
            DETACH DELETE r;
            """
        )
        self.write(
            """
            MATCH (r:Reaction)
            REMOVE r.canonicalId;
            """
        )

    def set_composite_reaction_labels(self):
        """
        Composite reaction nodes are labeled as ``Pathway`` and ``Reaction``.
        The ``Pathway`` label should be removed and a ``isCompositeReaction``
        property is added to the ``Reaction`` node.
        """
        self.write(
            """
            MATCH (n:Pathway:Reaction)
            REMOVE n:Pathway
            SET n.isCompositeReaction = true;
            """
        )

    def get_view_of_pathway(self, pathway_id: str):
        """
        Get the view of a pathway.

        Args:
            pathway_id: The pathway ID.

        Returns:
            The view of the pathway.
        """
        nodes = self.read(
            """
        MATCH (:Pathway {mcId: $pw_id})-[l:hasReaction]->(r:Reaction)
        WITH r
        MATCH (r)-[:is]->(rrdf:RDF),
              (r)-[:hasGeneProduct|hasComponent|hasMember*]->(gp:GeneProduct)-[:isEncodedBy]->(grdf:RDF)
        WITH r, rrdf.ecCode AS ec,
             COLLECT(gp.displayName) as symbol, COLLECT(grdf.ncbigene) AS ncbi
        RETURN {
          id: id(r), name: r.displayName, ec: ec, gibbs: r.gibbs0,
          direction: r.reactionDirection,
          genes: {symbol: symbol, ncbi: ncbi}
        };
            """,
            pw_id=pathway_id,
        )

        edges = self.read(
            """
        MATCH (:Pathway {mcId: $pw_id})-[l:hasReaction]->(r:Reaction)
        WITH COLLECT(r) AS nodes
        UNWIND nodes AS r1
        UNWIND nodes AS r2
        WITH * WHERE id(r1) < id(r2)
        MATCH p=(r1)-[:isPrecedingEvent]-(r2)
        UNWIND relationships(p) AS edge
        RETURN {
            id: id(edge), source: id(startNode(edge)), target: id(endNode(edge))
        };
            """,
            pw_id=pathway_id,
        )

        return nodes, edges

    def get_cpd_view_of_pathway(
        self,
        pathway_id: str,
        ignore_cpds: List[str] = COMMON_COMPOUNDS,
    ):
        """
        Get the view of a pathway with primary compounds linked by reactions.

        Args:
            pathway_id: The pathway mcId or displayName.
            ignore_cpds: A list of common compound displayNames to ignore.
        """
        res = self.read(
            """
            MATCH (:Pathway {displayName: $pw_id})-[:hasReaction]->(r:Reaction),
              (r)-[lr]->(cr:Compound),
              (r)-[lp]->(cp:Compound)
            WHERE (
              ($pw_id IN lr.isPrimaryReactantInPathway) OR
              (r.reactionDirection = 'reversible')
            ) AND (
              ($pw_id IN lp.isPrimaryProductInPathway) OR
              (r.reactionDirection = 'reversible')
            ) AND (
              id(lr) <> id(lp)
            ) AND (
              NOT cr.displayName IN $ignore_cpds
            ) AND (
              NOT cp.displayName IN $ignore_cpds
            )
            RETURN cr, cp,
              apoc.create.vRelationship(cr, r.displayName, properties(r), cp);
            """,
            pw_id=pathway_id,
            ignore_cpds=ignore_cpds,
        )

        return res



import atexit
import logging

from fastapi import FastAPI
from metabolike.api.config import db_conf
from metabolike.db.metacyc import MetaDB

logger = logging.getLogger(__name__)
app = FastAPI()

db = MetaDB(
    db_conf.metabolike_db_uri,
    db_conf.metabolike_db_user,
    db_conf.metabolike_db_password,
)
db.use_database(db_conf.metabolike_db_name)


@app.get("/")
def read_root():
    return {"Hello": "World"}


@app.get("/pathway/{pathway_id}")
def get_pathway(pathway_id: str):
    nodes, edges = db.get_view_of_pathway(pathway_id)
    return {"nodes": [x[0] for x in nodes], "edges": [x[0] for x in edges]}


def exit_application():
    db.close()


atexit.register(exit_application)



from typing import Optional

from pydantic import BaseModel


class DatabaseConfig(BaseModel):
    uri: str
    neo4j_user: str
    neo4j_password: str


class MetacycConfig(BaseModel):
    sbml: str
    reactions: Optional[str]
    atom_mapping: Optional[str]
    pathways: Optional[str]
    compounds: Optional[str]
    publications: Optional[str]
    classes: Optional[str]


class BrendaConfig(BaseModel):
    brenda_file: str
    bkms_react_file: str


class Config(BaseModel):
    database: DatabaseConfig
    metacyc: MetacycConfig
    brenda: Optional[BrendaConfig]



import logging
import re
from itertools import groupby
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

import libsbml
import pandas as pd
from metabolike.db.metacyc import MetaDB
from tqdm import tqdm

logger = logging.getLogger(__name__)

# MIRIAM qualifiers (https://co.mbine.org/standards/qualifiers)
# libsbml.BiolQualifierType_toString
BIO_QUALIFIERS = {
    0: "is",
    1: "hasPart",
    2: "isPartOf",
    3: "isVersionOf",
    4: "hasVersion",
    5: "isHomologTo",
    6: "isDescribedBy",
    7: "isEncodedBy",
    8: "encodes",
    9: "occursIn",
    10: "hasProperty",
    11: "isPropertyOf",
    12: "hasTaxon",
    13: "unknown",
}


class Metacyc:
    """
    Converting MetaCyc files to a Neo4j database.
    Documentation on the MetaCyc files and format FAQs can be found at:

    * MetaCyc data files download: https://metacyc.org/downloads.shtml
    * MetaCyc file formats: http://bioinformatics.ai.sri.com/ptools/flatfile-format.html
    * SBML FAQ: https://synonym.caltech.edu/documents/faq

    Args:
        neo4j: A :class:`.MetaDB` instance.
        sbml: The path to the MetaCyc SBML file to convert.
        reactions: The path to the ``reaction.dat`` file. If given,
            the file will be parsed and extra annotation on ``Reaction``
            nodes will be added.
        atom_mapping: The path to the ``atom-mappings-smiles.dat`` file. If
            given, the file will be parsed and chemical reactions in the ``SMILES``
            format will be added to the ``Reaction`` nodes.
        pathways: The path to the ``pathway.dat`` file. If given,
            the file will be parsed and pathway links will be added to the
            ``Reaction`` nodes.
        compounds: The path to the ``compound.dat`` file. If given, the
            file will be parsed and annotations on ``Compound`` nodes will
            be added.
        publications: The path to the ``publication.dat`` file. If given,
            the file will be parsed and annotations on ``Citation`` nodes
            will be added.
        classes: The path to the ``class.dat`` file. If given, the file
            will be parsed and annotations on ``Compartment``, ``Taxa``,
            and ``Compound`` nodes will be added.
        db_name: The name of the database to create. If None, the ``metaid``
            attribute of the SBML file is used.

    Attributes:
        db: A :class:`.MetaDB` instance. In :meth:`.setup`, this is connected
            to neo4j and used to perform all database operations. Should be
            closed after use.
        input_files: A dictionary of the paths to the input ``.dat`` files.
        missing_ids: A dictionary of sets of IDs that were not found in the
            input files. This is helpful for collecting IDs that appear to be
            in one class but are actually in another.
            :meth:`._report_missing_ids` can be used to print them out.
    """

    def __init__(
        self,
        neo4j: MetaDB,
        sbml: Union[str, Path],
        reactions: Optional[Union[str, Path]] = None,
        atom_mapping: Optional[Union[str, Path]] = None,
        pathways: Optional[Union[str, Path]] = None,
        compounds: Optional[Union[str, Path]] = None,
        publications: Optional[Union[str, Path]] = None,
        classes: Optional[Union[str, Path]] = None,
        db_name: Optional[str] = None,
    ):
        # Neo4j driver
        self.db = neo4j
        if db_name:
            self.db.use_database(db_name)

        # File paths
        self.input_files = {
            "sbml": validate_path(sbml),
            "reactions": validate_path(reactions),
            "atom_mapping": validate_path(atom_mapping),
            "pathways": validate_path(pathways),
            "compounds": validate_path(compounds),
            "publications": validate_path(publications),
            "classes": validate_path(classes),
        }
        logger.info(f"Input files: {self.input_files}")

        # Placeholder for missing IDs in the dat files
        self.missing_ids: Dict[str, Set[str]] = {
            "reactions": set(),
            "atom_mappings": set(),
            "pathways": set(),
            "compounds": set(),
            "publications": set(),
            "compartments": set(),
            "taxon": set(),
        }

    def setup(self, create_db: bool = True, **kwargs):
        """Enterpoint for setting up the database.

        The process is as follows:

        #. Parse the SBML file. All parsing errors are logged as warnings.
        #. If ``db_name`` is not given, use the ``metaid`` attribute of the SBML
           file to name the database.
        #. Create the database and constraints.
        #. Feed the SBML file into the database. This will populate
           ``Compartment``, ``Reaction``, ``Compound``, ``GeneProduct``,
           ``EntitySet``, ``Complex``, and ``RDF`` nodes.
        #. If ``reactions.dat`` is given, parse the file and add standard Gibbs
           free energy, standard reduction potential, reaction direction,
           reaction balance status, systematic name, comment attributes to
           ``Reaction`` nodes. Also link ``Reaction`` nodes to ``Pathway`` and
           ``Citation`` nodes.
           #. If ``pathways.dat`` is given:

              * Add synonyms, types, comments, common names to ``Pathway`` nodes.
              * Link ``Pathway`` nodes to super-pathway ``Pathway`` and taxonomy
                  ``Taxa`` nodes.
              * Link ``Reaction`` nodes within the pathway with
                  ``isPrecedingEvent`` relationships.
              * Link ``Pathway`` nodes with their rate limiting steps
                  (``Reaction`` nodes).
              * Link ``Pathway`` nodes with primary reactant and product
                  ``Compound`` nodes.
              * Link ``Pathway`` nodes with other ``Pathway`` nodes and annotate
                the shared ``Compound`` nodes.
              * For ``Reaction`` nodes, add ``isPrimary[Reactant|Product]InPathway``
                  labels in their links to ``Compound`` nodes.
        #. If ``atom-mappings-smiles.dat`` is given, parse the file and add
           SMILES_ mappings to ``Reaction`` nodes.

        #. If ``compounds.dat`` is given, parse the file and add standard Gibbs
           free energy, logP, molecular weight, monoisotopic molecular weight,
           polar surface area, pKa {1,2,3}, comment, and synonyms to ``Compound``
           nodes. Also add SMILES and INCHI strings to related ``RDF`` nodes, and
           link the ``Compound`` nodes to ``Citation`` nodes.
        #. If ``pubs.dat`` is given, parse the file and add DOI, PUBMED, MEDLINE IDs,
           title, source, year, URL, and ``REFERENT-FRAME`` to ``Citation`` nodes.
        #. If ``classes.dat`` is given, parse the file and:

           * Add common name and synonyms to ``Compartment`` nodes.
           * Add common name, strain name, comment, and synonyms to ``Taxa`` nodes.

        .. _SMILES: https://en.wikipedia.org/wiki/SMILES
        """
        # Read SBML file and set default database name
        doc = self._read_sbml(self.input_files["sbml"])
        model: libsbml.Model = doc.getModel()
        if not self.db.db_name:
            self.db.db_name = model.getMetaId().lower()

        # Setup Neo4j database and populate it with data
        self.db.use_database(self.db.db_name)
        self.db.setup_graph_db(create_db=create_db, **kwargs)
        self.sbml_to_graph(model)

        # Add additional information of reactions to the graph if given
        if self.input_files["reactions"] or self.input_files["atom_mapping"]:
            all_rxns = self.db.get_all_nodes("Reaction", "displayName")

            if self.input_files["reactions"]:
                logger.info("Adding additional reaction information to the graph")
                rxn_dat = self._read_dat_file(self.input_files["reactions"])

                for rxn in tqdm(all_rxns, desc="reactions.dat file"):
                    self.reaction_to_graph(rxn, rxn_dat)
                    logger.debug(f"Added extra info for reaction {rxn}")

                # Read pathways file if given
                if self.input_files["pathways"]:
                    logger.info("Creating pathway links")
                    pw_dat = self._read_dat_file(self.input_files["pathways"])

                    all_pws = self.db.get_all_nodes("Pathway", "mcId")
                    self.super_pathways = set()
                    for pw in tqdm(all_pws, desc="pathways.dat file"):
                        self.pathway_to_graph(pw, pw_dat, rxn_dat)
                        logger.debug(f"Added pathway annotation for {pw}")

                    # Annotate the super-pathways
                    self.super_pathways = self.super_pathways - set(all_pws)
                    while self.super_pathways:
                        spw = self.super_pathways.pop()
                        self.pathway_to_graph(spw, pw_dat, rxn_dat)
                        logger.debug(f"Added pathway annotation for {spw}")

                    del self.super_pathways

                    # Correct composite reactions
                    self.db.set_composite_reaction_labels()

                # Fix the direction of reactions
                self.db.fix_reaction_direction()

            # SMILES reactions
            if self.input_files["atom_mapping"]:
                logger.info("Adding SMILES to reactions")
                smiles = pd.read_table(
                    self.input_files["atom_mapping"],
                    sep="\t",
                    header=None,
                    names=["rxn", "smiles"],
                )
                smiles: Dict[str, str] = smiles.set_index("rxn").to_dict()["smiles"]
                for rxn in tqdm(all_rxns, desc="atom_mapping.dat file"):
                    self.atom_mapping_to_graph(rxn, smiles)

        # Compounds in compounds.dat
        if self.input_files["compounds"]:
            logger.info("Annotating Compound nodes")
            cpd_dat = self._read_dat_file(self.input_files["compounds"])

            all_cpds = self.db.get_all_compounds()
            for cpd, biocyc in tqdm(all_cpds, desc="compounds.dat file"):
                self.compounds_to_graph(cpd, biocyc, cpd_dat)
                logger.debug(f"Added annotation for compound {cpd} with {biocyc}")

        # Read publications file if given
        if self.input_files["publications"]:
            logger.info("Annotating publications")
            pub_dat = self._read_dat_file(self.input_files["publications"])

            all_cits = self.db.get_all_nodes("Citation", "mcId")
            for cit in tqdm(all_cits, desc="pubs.dat file"):
                self.citation_to_graph(cit, pub_dat)
                logger.debug(f"Added annotation for citation {cit}")

        # Compartments, Taxon, and comments of Compounds in classes.dat
        if self.input_files["classes"]:
            logger.info("Adding common names from classes.dat")
            class_dat = self._read_dat_file(self.input_files["classes"])
            self.classes_to_graph(class_dat)

        self._report_missing_ids()

    def sbml_to_graph(self, model: libsbml.Model):
        """
        Populate Neo4j database with SBML data.

        Nodes are created for each SBML element using ``MERGE`` statements:
        https://neo4j.com/docs/cypher-manual/current/clauses/merge/#merge-merge-with-on-create
        """
        # Compartments
        logger.info("Creating Compartment nodes")
        for c in model.getListOfCompartments():
            c: libsbml.Compartment
            props = {"displayName": c.getName()}
            self.db.create_node("Compartment", c.getMetaId(), props)

        # Compounds, i.e. metabolites, species
        logger.info("Creating Compound nodes")
        for s in tqdm(model.getListOfSpecies(), desc="Compounds"):
            s: libsbml.Species
            # Basic properties
            mcid: str = s.getMetaId()
            props = {
                "displayName": s.getName(),
                "charge": s.getCharge(),
                "chemicalFormula": s.getPlugin("fbc").getChemicalFormula(),
                "boundaryCondition": s.getBoundaryCondition(),  # unchanged by reactions
                "hasOnlySubstanceUnits": s.getHasOnlySubstanceUnits(),
                "constant": s.getConstant(),
            }
            self.db.create_compound_node(s.getCompartment(), mcid, props)

            # Add RDF annotations
            logger.debug(f"Adding RDF nodes for Compound {mcid}")
            cvterms: List[libsbml.CVTerm] = s.getCVTerms()
            for cvterm in cvterms:
                self._add_sbml_rdf_node("Compound", mcid, cvterm)

        # Gene products
        logger.info("Creating GeneProduct nodes")
        fbc: libsbml.FbcModelPlugin = model.getPlugin("fbc")
        for gp in tqdm(fbc.getListOfGeneProducts(), desc="GeneProducts"):
            gp: libsbml.GeneProduct
            mcid = gp.getMetaId()
            props = {
                "displayName": gp.getName(),
                "label": gp.getLabel(),
            }
            self.db.create_node("GeneProduct", mcid, props)

            logger.debug(f"Adding RDF nodes for GeneProduct {mcid}")
            cvterms: List[libsbml.CVTerm] = gp.getCVTerms()
            for cvterm in cvterms:
                self._add_sbml_rdf_node("GeneProduct", mcid, cvterm)

        # Reactions
        logger.info("Creating Reaction nodes")
        for r in tqdm(model.getListOfReactions(), desc="Reactions"):
            r: libsbml.Reaction
            mcid = r.getMetaId()

            # Basic properties
            props = {
                "displayName": r.getName(),
                "fast": r.getFast(),
            }
            self.db.create_node("Reaction", mcid, props)

            # Add RDF nodes
            logger.debug(f"Adding RDF nodes for Reaction {mcid}")
            cvterms: List[libsbml.CVTerm] = r.getCVTerms()
            for cvterm in cvterms:
                self._add_sbml_rdf_node("Reaction", mcid, cvterm)

            # Add reactants and products
            logger.debug(f"Adding reactants for Reaction {mcid}")
            reactants = r.getListOfReactants()
            self._link_reaction_to_compound(mcid, reactants, "Left")

            logger.debug(f"Adding products for Reaction {mcid}")
            products = r.getListOfProducts()
            self._link_reaction_to_compound(mcid, products, "Right")

            # Add associated gene products
            # This could be complicated where the child nodes could be:
            # 1. GeneProductRef
            # 2. fbc:or -> GeneProductRef
            # 3. fbc:and -> GeneProductRef
            # 4. Arbitrarily nested fbc:or / fbc:and within cases 2 and 3
            gpa: libsbml.GeneProductAssociation = r.getPlugin(
                "fbc"
            ).getGeneProductAssociation()
            if gpa is not None:
                node = gpa.getAssociation()
            self._add_sbml_gene_product_association_node(node, mcid)

    def reaction_to_graph(self, rxn_id: str, rxn_dat: Dict[str, List[List[str]]]):
        """
        Parse one entry from the reaction attribute-value file, and add relevant
        information to the graph database in one transaction.

        Args:
            rxn_id: The *full* reaction ID from the graph database.
            rxn_dat: The reactions.dat file as an attribute-value list.
        """
        canonical_id = self._find_rxn_canonical_id(rxn_id, rxn_dat.keys())
        if canonical_id not in rxn_dat:
            self.missing_ids["reactions"].add(canonical_id)
            return
        lines = rxn_dat[canonical_id]
        props: Dict[str, Union[str, List[str]]] = {"canonicalId": canonical_id}

        for k, v in lines:
            # SYNONYMS is a special case because it is a list
            if k in {"SYNONYMS", "TYPES"}:
                _add_kv_to_dict(props, k, v, as_list=True)
            elif k in {
                "GIBBS-0",
                "STD-REDUCTION-POTENTIAL",
                "REACTION-DIRECTION",
                "REACTION-BALANCE-STATUS",
                "SYSTEMATIC-NAME",
                "COMMENT",  # TODO: link to other nodes
            }:
                _add_kv_to_dict(props, k, v, as_list=False)
            elif k == "IN-PATHWAY":
                self.db.link_reaction_to_pathway(rxn_id, v)
            elif k == "RXN-LOCATIONS":
                self.db.link_reaction_to_compartment(rxn_id, v)
            elif k == "CITATIONS":
                self.db.link_node_to_citation("Reaction", rxn_id, v)

        # Clean up props before writing to graph
        props = self._clean_props(
            props,
            num_fields=[
                _snake_to_camel(x) for x in ["GIBBS-0", "STD-REDUCTION-POTENTIAL"]
            ],
            enum_fields=[
                _snake_to_camel(x)
                for x in ["REACTION-BALANCE-STATUS", "REACTION-DIRECTION"]
            ],
        )
        self.db.add_props_to_node("Reaction", "displayName", rxn_id, props)

    def atom_mapping_to_graph(self, rxn_id: str, smiles: Dict[str, str]):
        """
        Args:
            rxn_id: The *full* reaction ID from the graph database.
            smiles: The reaction ID -> SMILES dictionary from the atom mapping file.
        """
        canonical_id = self._find_rxn_canonical_id(rxn_id, smiles.keys())
        if canonical_id not in smiles:
            self.missing_ids["atom_mappings"].add(canonical_id)
            return
        props = {"smiles_atom_mapping": smiles[canonical_id]}
        self.db.add_props_to_node("Reaction", "displayName", rxn_id, props)

    def pathway_to_graph(
        self,
        pw_id: str,
        pw_dat: Dict[str, List[List[str]]],
        rxn_dat: Dict[str, List[List[str]]],
    ):
        """
        Parse one entry from the pathway attribute-value file, and add relevant
        information to the graph database in one transaction.

        Args:
            pw_id: The pathway ID from the graph database.
            pw_dat: The pathways.dat file as an attribute-value list.
            rxn_dat: The reactions.dat file as an attribute-value list. This is
                used to find the annotation data for composite reactions marked
                as pathways.
        """
        if (pw_id not in pw_dat) and (pw_id not in rxn_dat):
            self.missing_ids["pathways"].add(pw_id)
            return

        # Deal with pathway nodes with reaction IDs (composite reactions)
        if pw_id in rxn_dat:
            self.reaction_to_graph(pw_id, rxn_dat)
            # Merge the Reaction node with the Pathway node under the same ID
            self.db.merge_nodes("Pathway", "Reaction", "mcId", "displayName", pw_id)
            return

        lines = pw_dat[pw_id]
        props: Dict[str, Union[str, List[str]]] = {"displayName": pw_id}
        for k, v in lines:
            # Pathway node properties
            if k in {"SYNONYMS", "TYPES"}:
                _add_kv_to_dict(props, k, v, as_list=True)
            elif k in {"COMMENT", "COMMON-NAME"}:
                _add_kv_to_dict(props, k, v, as_list=False)

            # Relationship with other nodes
            elif k in {"IN-PATHWAY", "SUPER-PATHWAYS"}:
                self.super_pathways.add(v)
                self.db.link_pathway_to_superpathway(pw_id, v)
            elif k == "CITATIONS":
                self.db.link_node_to_citation("Pathway", pw_id, v)
            elif k == "SPECIES":
                self.db.link_pathway_to_taxa(pw_id, v, "hasRelatedSpecies")
            elif k == "TAXONOMIC-RANGE":
                self.db.link_pathway_to_taxa(pw_id, v, "hasExpectedTaxonRange")
            elif k == "PREDECESSORS":
                rxns = v[2:-2].split('" "')  # leading (" and trailing ")
                r1, r2 = rxns[0], rxns[1:]  # could have multiple predecessors
                self.db.link_reaction_to_next_step(r1, r2, pw_id)
            elif k == "RATE-LIMITING-STEP":
                self.db.link_pathway_to_rate_limiting_step(pw_id, v)
            elif k in {"PRIMARY-PRODUCTS", "PRIMARY-REACTANTS"}:
                relationship = "Product" if k == "PRIMARY-PRODUCTS" else "Reactant"
                self.db.link_pathway_to_primary_compound(pw_id, v, relationship)
            elif k == "REACTION-LAYOUT":
                rxn_id, d = self._parse_reaction_layout(v)
                if not rxn_id:
                    continue
                # Two cypher statements for Reactant and Product compounds
                for side, compounds in d.items():
                    self.db.link_reaction_to_primary_compound(
                        rxn_id, compounds, pw_id, side
                    )
            elif k == "PATHWAY-LINKS":
                cpd_id, pw_ids, direction = self._parse_pathway_links(v)
                if pw_ids:
                    self.db.link_pathway_to_pathway(pw_id, pw_ids, direction, cpd_id)

        # Write Pathway node properties
        self.db.add_props_to_node("Pathway", "mcId", pw_id, props)

    def compounds_to_graph(
        self,
        cpd: str,
        biocyc: str,
        cpd_dat: Dict[str, List[List[str]]],
    ):
        """Annotate a compound node with data from the compound.dat file.

        Args:
            cpd: The ``displayName`` of the compound.
            biocyc: The biocyc id of the compound.
            cpd_dat: The compound.dat data.
            session: The neo4j session.
        """
        if biocyc not in cpd_dat:
            self.missing_ids["compounds"].add(biocyc)
            return
        lines = cpd_dat[biocyc]
        c_props: Dict[str, Union[str, List[str]]] = {}
        rdf_props: Dict[str, Union[str, List[str]]] = {}
        for k, v in lines:
            if k in {
                "COMMON-NAME",
                "GIBBS-0",
                "LOGP",
                "MOLECULAR-WEIGHT",
                "MONOISOTOPIC-MW",
                "POLAR-SURFACE-AREA",
                "PKA1",
                "PKA2",
                "PKA3",
                "COMMENT",
            }:
                _add_kv_to_dict(c_props, k, v, as_list=False)
            elif k == "SYNONYMS":
                _add_kv_to_dict(c_props, k, v, as_list=True)
            elif k in {"SMILES", "INCHI"}:
                _add_kv_to_dict(rdf_props, k, v, as_list=False)
            elif k == "DBLINKS":
                pass  # TODO: parse DBLINKS
            elif k == "CITATIONS":
                self.db.link_node_to_citation("Compound", cpd, v)

        c_props = self._clean_props(
            c_props,
            num_fields=[
                _snake_to_camel(x)
                for x in [
                    "GIBBS-0",
                    "LOGP",
                    "MOLECULAR-WEIGHT",
                    "MONOISOTOPIC-MW",
                    "POLAR-SURFACE-AREA",
                    "PKA1",
                    "PKA2",
                    "PKA3",
                ]
            ],
            enum_fields=[],
        )
        self.db.add_props_to_compound(cpd, c_props, rdf_props)

    def citation_to_graph(self, cit_id: str, pub_dat: Dict[str, List[List[str]]]):
        """Annotate a citation node with data from the publication.dat file.

        If there are multiple fields in the given ``cit_id``, then the fields are
        separated by colons. The first field is the citation ID, the second is the
        evidence type (in classes.dat), the third is not documented, and the
        fourth is the curator's name.

        In most cases the citation ID should match: ``PUB-[A-Z0-9]+$``,
        with a few exceptions containing double dashes, e.g. ``PUB--8``,
        or some dashes within author names, e.g. ``PUB-CHIH-CHING95``.

        Args:
            cit_id: The citation ``mcId`` property.
            pub_dat: The publication.dat data.
        """
        # TODO: deal with evidence frames. Evidence frames are in the form of
        # 10066805:EV-EXP-IDA:3354997111:hartmut
        pub_dat_id = re.sub(r"[\[\]\s,']", "", cit_id.split(":")[0].upper())
        pub_dat_id = re.sub(r"-(\d+)", r"\1", pub_dat_id)
        if pub_dat_id == "BOREJSZA-WYSOCKI94":
            pub_dat_id = "BOREJSZAWYSOCKI94"  # only exception with dash removed
        if not pub_dat_id:
            return
        pub_dat_id = "PUB-" + pub_dat_id
        if pub_dat_id not in pub_dat:
            self.missing_ids["publications"].add(pub_dat_id)
            return

        lines = pub_dat[pub_dat_id]
        props: Dict[str, Union[str, List[str]]] = {"citationId": pub_dat_id}
        for k, v in lines:
            # Pathway node properties
            if k == "AUTHORS":
                _add_kv_to_dict(props, k, v, as_list=True)
            elif k in {
                "DOI-ID",
                "PUBMED-ID",
                "MEDLINE-ID",
                "TITLE",
                "SOURCE",
                "YEAR",
                "URL",
                "REFERENT-FRAME",
            }:
                _add_kv_to_dict(props, k, v, as_list=False)

        self.db.add_props_to_node("Citation", "mcId", cit_id, props)

    def classes_to_graph(self, class_dat: Dict[str, List[List[str]]]):
        """Parse the classes.dat file.

        Add common name and synonyms to ``Compartment`` nodes. Add common name,
        strain name, comment, and synonyms to ``Taxa`` nodes.

        Args:
            class_dat: The classes.dat data.
        """
        # Common names for cell components
        all_cco = self.db.get_all_nodes("Compartment", "displayName")
        for cco in all_cco:
            if cco not in class_dat:
                self.missing_ids["compartments"].add(cco)
                continue
            props: Dict[str, Union[str, List[str]]] = {}
            for k, v in class_dat[cco]:
                if k == "COMMON-NAME":
                    _add_kv_to_dict(props, k, v, as_list=False)
                elif k == "SYNONYMS":
                    _add_kv_to_dict(props, k, v, as_list=True)

            self.db.add_props_to_node("Compartment", "displayName", cco, props)

        # Common names and synonyms for organisms. Some also have strain names
        all_taxon = self.db.get_all_nodes("Taxa", "mcId")
        for taxa in tqdm(all_taxon, desc="Taxon in classes.dat"):
            if taxa not in class_dat:
                self.missing_ids["taxon"].add(taxa)
                continue
            props: Dict[str, Union[str, List[str]]] = {}
            for k, v in class_dat[taxa]:
                if k in {"COMMON-NAME", "STRAIN-NAME", "COMMENT"}:
                    _add_kv_to_dict(props, k, v, as_list=False)
                elif k == "SYNONYMS":
                    _add_kv_to_dict(props, k, v, as_list=True)
                # TODO: TYPES links the taxon

            self.db.add_props_to_node("Taxa", "mcId", taxa, props)

        # TODO: Evidence code in citations are in the `Evidence` attr

    @staticmethod
    def _read_sbml(sbml_file: Path) -> libsbml.SBMLDocument:
        reader = libsbml.SBMLReader()
        metacyc = reader.readSBMLFromFile(sbml_file)
        logger.info("Finished reading SBML file")

        for i in range(metacyc.getNumErrors()):
            err = metacyc.getError(i)
            logger.warning(
                "SBML reader raised %s during parsing at line %d, column %d: %s",
                err.getSeverityAsString(),
                err.getLine(),
                err.getColumn(),
                err.getMessage().replace("\n", " "),
            )

        return metacyc

    def _add_sbml_rdf_node(self, node_label: str, mcid: str, cvterm: libsbml.CVTerm):
        """Create RDF node and link it to the given SBML node.

        RDF in Annotation are in the form of triples:
        the model component to annotate (subject), the relationship between
        the model component and the annotation (predicate), and a term
        describing the component (object).

        Args:
            node_label: The label of the SBML node to annotate. Should be one
                of the labels defined in ``NODE_LABELS``.
            mcid: The MetaCyc ID of the SBML node to annotate.
            cvterm: The CVTerm that contains information to add to the RDF node.
        """
        # Get the biological qualifier type of the terms
        bio_qual = BIO_QUALIFIERS[cvterm.getBiologicalQualifierType()]
        # Get the content of each RDF term
        uris = [
            self._split_uri(cvterm.getResourceURI(i))
            for i in range(cvterm.getNumResources())
        ]

        props: Dict[str, Union[str, List[str]]] = {}
        for resource, identifier in uris:
            is_list_resource = resource == "ec-code"
            _add_kv_to_dict(props, resource, identifier, as_list=is_list_resource)

        self.db.link_node_to_rdf(node_label, mcid, bio_qual, props)

    def _link_reaction_to_compound(
        self,
        reaction_id: str,
        compounds: List[libsbml.SpeciesReference],
        compound_type: str,
    ):
        """Link reactants or products to a reaction.

        Args:
            reaction_id: The MetaCyc ID of the reaction.
            compounds: The list of compounds to link to the reaction.
            compound_type: The type of compound to link to the reaction. Should
                be one of "Left" or "Right".
        """
        if compound_type not in ["Left", "Right"]:
            raise ValueError(f"Invalid compound type: {compound_type}")
        for cpd in compounds:
            props = {
                "stoichiometry": cpd.getStoichiometry(),
                "constant": cpd.getConstant(),
            }
            self.db.link_reaction_to_compound(
                reaction_id, cpd.getSpecies(), compound_type, props
            )

    def _add_sbml_gene_product_association_node(
        self,
        node: Union[libsbml.GeneProductRef, libsbml.FbcAnd, libsbml.FbcOr],
        source_id: str,
        source_label: str = "Reaction",
        edge_type: str = "hasGeneProduct",
        node_index: int = 0,
    ):
        """
        Add gene products to a reaction. When the added node is FbcAnd or FbcOr,
        recursively add the children. This means a custom ``mcId`` is constructed
        for the ``Complex`` and ``EntitySet`` nodes corresponding to the ``FbcAnd``
        and ``FbcOr`` nodes, respectively.

        Args:
            node: The GeneProductAssociation child node to add.
            source_id: The MetaCyc ID of the source node. This should be the
                MetaCyc ID of the ``Reaction`` node
            source_label: The label of the source node.
            edge_type: The type of edge to add. Should be one of ``hasGeneProduct``,
                ``hasComponent``, or ``hasMember``.
            node_index: The index of the current node. This is used to construct
                the ``mcId`` of the ``Complex`` and ``EntitySet`` nodes.
        """
        # If there's no nested association, add the node directly
        if isinstance(node, libsbml.GeneProductRef):
            self.db.link_node_to_gene(
                source_label, source_id, node.getGeneProduct(), "GeneProduct", edge_type
            )

        # For nested associations, first add a `Complex` or `EntitySet` node,
        # then recursively add the children
        elif isinstance(node, libsbml.FbcAnd):
            complex_id = f"{source_id}_complex{node_index}"
            self.db.link_node_to_gene(
                source_label, source_id, complex_id, "Complex", edge_type
            )

            for i in range(node.getNumAssociations()):
                self._add_sbml_gene_product_association_node(
                    node.getAssociation(i),
                    complex_id,
                    "Complex",
                    "hasComponent",
                    i,
                )
        elif isinstance(node, libsbml.FbcOr):
            eset_id = f"{source_id}_entityset{node_index}"
            self.db.link_node_to_gene(
                source_label, source_id, eset_id, "EntitySet", edge_type
            )

            for i in range(node.getNumAssociations()):
                self._add_sbml_gene_product_association_node(
                    node.getAssociation(i),
                    eset_id,
                    "EntitySet",
                    "hasMember",
                    i,
                )
        else:
            logging.error(f"Unhandled GeneProductAssociation type {type(node)}")
            raise ValueError

    @staticmethod
    def _split_uri(uri: str) -> Tuple[str, str]:
        """Split a URI into a namespace and an annotation term.

        Args:
            uri: URI to split.

        Returns:
            Tuple of namespace and annotation term.
        """
        # First three elements are from http://identifiers.org/
        res = uri.split("/")[3:]
        resource, identifier = res[0], res[1:]
        resource = resource.replace(".", "-")  # Ec-code
        identifier = "".join(identifier)

        # In some cases the identifier in the RDF nodes of the SBML file has a
        # prefix, e.g. META: or HUMAN: for BioCyc IDs. We don't want these as
        # they make later queries difficult.
        identifier = re.sub(r"^[A-Z]+:", "", identifier)
        return resource, identifier

    @staticmethod
    def _read_dat_file(filepath: Union[str, Path]) -> Dict[str, List[List[str]]]:
        # `rb` to bypass the 0xa9 character
        with open(filepath, "rb") as f:
            lines = [l.decode("utf-8", "ignore").strip() for l in f]
            # Also remove empty lines and comments on the top of the file
            lines = [l for l in lines if l and (not l.startswith("#"))]

        # Split entries based on `//`
        docs: Dict[str, List[List[str]]] = {}
        for k, g in groupby(lines, key=lambda x: x != "//"):
            if not k:
                continue

            doc = list(g)
            # Concatenate attributes with multiple lines (mostly COMMENT)
            doc_txt = "\n".join(doc)
            doc_txt = doc_txt.replace("\n/", " ")
            doc = doc_txt.split("\n")

            # Split key-attribute pairs
            doc = [l.split(" - ", maxsplit=1) for l in doc]
            uniq_id = doc[0][1]

            # Remove empty values like "SYNONYMS - " in ORG-6026
            doc = [x for x in doc[1:] if len(x) == 2]

            docs[uniq_id] = doc

        return docs

    @staticmethod
    def _find_rxn_canonical_id(rxn_id: str, all_ids: Iterable[str]) -> str:
        """Find the canonical ID for a reaction.

        Some reactions have longer ID forms in ``metabolic-reactions.xml`` than
        in ``reactions.dat`` or ``pathways.dat``. For example,
        ``F16BDEPHOS-RXN`` has two counterparts in ``metabolic-reactions.xml``:
        ``F16BDEPHOS-RXN[CCO-PERI-BAC]-FRUCTOSE-16-DIPHOSPHATE/WATER//FRUCTOSE-6P/Pi.60.``
        and
        ``F16BDEPHOS-RXN[CCO-CYTOSOL]-FRUCTOSE-16-DIPHOSPHATE/WATER//FRUCTOSE-6P/Pi.59.``

        This helper function extracts the leading part from the full ID.

        Args:
            rxn_id: The MetaCyc ID of the reaction.
            all_ids: All UNIQUE-IDs in the reactions.dat file.

        Returns:
            The canonical ID for the reaction.
        """
        # See if the rxn_id is in reaction.dat
        if rxn_id in all_ids:
            return rxn_id

        # If not, extract the ID from the leading part of rxn_id
        match_canonical_id = re.match(
            r"((TRANS-)?(RXN[A-Z\d]*)-\d+)|([A-Z\d.\-\+]+RXN)", rxn_id
        )
        if match_canonical_id:
            canonical_id = match_canonical_id.group(0)
            return canonical_id
        else:
            raise ValueError(f"rxn_id has no canonical form: {rxn_id}")

    @staticmethod
    def _clean_props(
        props: Dict[str, Any], num_fields: Iterable[str], enum_fields: Iterable[str]
    ) -> Dict[str, Any]:
        """Normalize properties to be used in Cypher.

        Args:
            props: Properties to normalize.
            num_fields: Fields that should be converted to float numbers.
            enum_fields: Fields that should be converted to alphanumerical strings.

        Returns:
            A dictionary with normalized properties.
        """
        for f in num_fields:
            if f in props:
                props[f] = float(props[f])

        enum_pattern = re.compile(r"\W+")
        for f in enum_fields:
            if f in props:
                props[f] = props[f].replace("-", "_").lower()
                props[f] = enum_pattern.sub("", props[f])

        return props

    @staticmethod
    def _parse_reaction_layout(s: str) -> Tuple[str, Dict[str, List[str]]]:
        """Parse the reaction layout from the ``reactions.dat`` file.

        The attribute has the following format: ``(<rxn-id> (:LEFT-PRIMARIES
        <cpd-id> ...) (:DIRECTION :[L2R|R2L]) (:RIGHT-PRIMARIES <cpd-id> ...))``

        Args:
            s: The layout string.

        Returns:
            The reaction canonical ID, and a dictionary with the parsed layout
            containing the following keys: - ``Reactant``: List of primary
            reactant compounds. - ``Product``: List of primary product
            compounds.
        """
        m = re.compile(
            r"\(([^\(]+) "  # reaction ID
            r"\(:LEFT-PRIMARIES ([^\)]+)\) "  # left primary compounds
            r"\(:DIRECTION :(L2R|R2L)\) "  # reaction direction
            r"\(:RIGHT-PRIMARIES ([^\)]+)\)\)"
        )
        res = m.fullmatch(s)
        if not res:
            # Sometimes there's no left/right primaries and only a direction
            # when the rxn_id is actually a pathway ID
            return "", {}

        rxn_id, l_prim, direction, r_prim = res.groups()
        # Use direction to determine which side are the reactants
        if direction == "L2R":
            reactants = l_prim.split(" ")
            products = r_prim.split(" ")
        else:
            reactants = r_prim.split(" ")
            products = l_prim.split(" ")

        d = {
            "Reactant": reactants,
            "Product": products,
        }

        return rxn_id, d

    @staticmethod
    def _parse_pathway_links(s: str) -> Tuple[str, List[str], str]:
        """
        Parse ``PATHWAY-LINKS`` from the ``pathways.dat`` file. This connects
        the ``Pathway`` nodes in the graph through shared ``Compound`` nodes.
        The format of the input string has several possibilities. The simplest
        is when the pathway points to one or more target pathways:

            (<compound-id> <target-pathway-ids ...>)

        These ``target-pathway-ids`` are inheriently *outgoing* links. Sometimes
        the direction is explicitly specified. When there are multiple pathway
        IDs and only one is specified as ``OUTGOING``, all the other pathways
        are also *outgoing* links.

            (<compound-id> (<target-pathway-id> . :OUTGOING))
            (<compound-id> (<target-pathway-id> . :INCOMING))

        The compound or pathway ID could also be a frame ID instead of the unique
        ID of the object, in which case the ID is wrapped in pipe symbols. This
        mostly happens for compound IDs, and when it happens for pathway IDs it's
        often a pathway type instead of the unique ID.

            (|<compound-frame-id>| <target-pathway-id>)

        Finally, the target pathway ID could be some random string wrapped in
        quotes. We can safely ignore most of these (e.g. "dietary input",
        "release from the lysosome"), but in several cases valid pathway IDs
        are also wrapped in quotes.

        Args:
            s: The pathway links string to parse.

        Returns:
            A tuple with the first element being the compound ID, the second
            element being a list of pathway IDs, and the third element being
            the direction of the link. The list of pathway IDs could be empty
            because non-canonical pathway IDs are dropped.
        """
        # Extract the compound ID at the beginning of the string
        cpd_id_rgx = re.compile(r"^\(([^\s]+) ")
        res = cpd_id_rgx.match(s)
        if not res:
            logging.warning(f"Pathway links string doesn't have a compound: {s}")
            return "", [], ""
        cpd = res.group(1)
        s = s[res.end() : -1]  # remove cpd ID and closing parenthesis

        # Extract pathway links recursively
        pathways: List[str] = []
        direction = ""
        pw_id = r"\"[^\"]+\"|[^\s]+"
        directed_pw_rgx = re.compile(rf"\(({pw_id}) \. :(INCOMING|OUTGOING)\)")
        pw_rgx = re.compile(pw_id)

        while s:
            # Directed pathway links wrapped in parentheses
            if s[0] == "(":
                m = directed_pw_rgx.match(s)
                pathways.append(m.group(1))
                direction = m.group(2)
            # Regular pathway links separated by spaces
            else:
                m = pw_rgx.match(s)
                pathways.append(m.group(0))
            s = s[m.end() :].strip()

        # Final cleanup
        cpd = cpd.replace("|", "")
        pathways = [
            p.replace('"', "") for p in pathways if (" " not in p) and p.isupper()
        ]  # remove frame IDs and quoted annotations
        direction = "INCOMING" if direction == "INCOMING" else "OUTGOING"
        return cpd, pathways, direction

    def _report_missing_ids(self):
        for datfile, ids in self.missing_ids.items():
            if ids:
                logger.warning(f"The following IDs were not found in {datfile}: {ids}")


def validate_path(filepath: Optional[Union[str, Path]]) -> Optional[Path]:
    if not filepath:
        return None
    f = Path(filepath).expanduser().resolve()
    if not f.is_file():
        logger.error(f"File does not exist: {f}")
        raise FileNotFoundError(str(f))

    return f


def _snake_to_camel(s: str, sep: str = "-") -> str:
    """Convert snake_case to CamelCase.

    Args:
        s: String to convert.
        sep: Separator to use.

    Returns:
        camelCase string.
    """
    s_camel = [x.capitalize() for x in s.split(sep)]
    s_camel[0] = s_camel[0].lower()
    return "".join(s_camel)


def _add_kv_to_dict(
    d: Dict[str, Any], k: str, v: Any, as_list: bool = False
) -> Dict[str, Any]:
    """Add a key-value pair to a dictionary.

    Args:
        d: Dictionary to add the key-value pair to.
        k: Key to add.
        v: Value to add.

    Returns:
        Dictionary with the key-value pair added. If the key already exists,
        the value is saved as a list.
    """
    k_camel = _snake_to_camel(k)
    if as_list:
        d.setdefault(k_camel, []).append(v)
    else:
        d[k_camel] = v

    return d



import tarfile
from io import StringIO
from pathlib import Path

import pandas as pd
import requests


def get_bkms_tarball(filepath: str, extract: bool = True) -> None:
    """
    The file is available at: https://bkms.brenda-enzymes.org/download.php

    The compressed file (``Reactions_BKMS.tar.gz``) includes the table in tab
    stop separated format (Excel, OpenOffice). The table contains actual data of
    BRENDA (release 2021.2, only reactions with naturally occuring substrates),
    MetaCyc (version 24.5), SABIO-RK (07/02/2021) and KEGG data, downloaded on
    the 23th of April 2012. Downloading more recent KEGG data cannot be offered
    because a KEGG license agreement would be necessary.

    Args:
        filepath: The path to store the downloaded file.
        extract: Extract the file.
    """
    url = "https://bkms.brenda-enzymes.org/download/Reactions_BKMS.tar.gz"
    r = requests.get(url, stream=True)
    with open(filepath, "wb") as f:
        for chunk in r.iter_content(chunk_size=1024):
            f.write(chunk)

    if extract:
        _extract_bkms_tarball(filepath)


def _extract_bkms_tarball(filepath: str) -> None:
    """
    Extract the tsv file from the tarball.

    Args:
        filepath: The path to the downloaded file.
    """
    tar = tarfile.open(filepath, "r:gz")
    tar.extractall()
    tar.close()


def read_bkms(filepath: str, clean: bool = True) -> pd.DataFrame:
    """
    Read the BKMS-react table and prepare it for further processing.

    The table contains random ``^M`` characters in some rows. These characters
    won't break pandas, but they will make the parsed table unexpectedly long.
    The ``clean`` parameter can be used to remove these characters.

    BRENDA takes EC numbers as identifiers, so we need entries with non-empty
    ``EC_Number`` and ``Reaction_ID_MetaCyc`` columns. There is a
    ``Reaction_ID_BRENDA`` column that's sometimes non-empty when the EC number
    is missing, but the field is not documented and it's not clear how it can be
    mapped to an entry in the BRENDA text file.

    To be extra conservative, we only keep entries with non-empty
    ``Reaction_ID_KEGG`` columns. Only reactions in MetaCyc/BioCyc that are
    associated with matching EC numbers *and* KEGG IDs will be annotated.

    Args:
        filepath: The path to the BKMS-react ``.tsv`` file.
        clean: Remove the``^M`` characters.

    Returns:
        A pandas dataframe:
    """
    fp = Path(filepath).expanduser().resolve()
    if not clean:
        df = pd.read_table(fp, sep="\t")
        return df.drop_duplicates()

    with open(fp, "r") as f:
        lines = f.readlines()
    lines = [line.replace(r"\r", "") for line in lines]
    df: pd.DataFrame = pd.read_table(StringIO("".join(lines)), sep="\t")
    df = (
        df.filter(
            [
                "EC_Number",
                "Reaction_ID_KEGG",
                "Reaction_ID_MetaCyc",
                "MetaCyc_Pathway_ID",
                "Reaction",
            ]
        )
        .query("Reaction_ID_MetaCyc.notna() & Reaction_ID_KEGG.notna()")
        .drop_duplicates()
    )
    # TODO: what are the BRENDA reaction IDs? Can we map them to MetaCyc?
    # TODO: Can the MetaCyc IDs be mapped to organism-specific BioCyc IDs?
    # TODO: clean up the EC numbers (non-canonical, missing values, etc.)
    return df



#!/usr/bin/env python
import logging
from pathlib import Path
from typing import Optional

import typer
import uvicorn
import yaml

from metabolike.config import Config
from metabolike.db.metacyc import MetaDB
from metabolike.parser.brenda import parse_brenda
from metabolike.parser.metacyc import Metacyc

logging.basicConfig(
    format="[%(levelname)s] %(asctime)s - %(name)s:%(lineno)s:%(funcName)s - %(message)s",
    level=logging.INFO,
)
logger = logging.getLogger(__name__)

app = typer.Typer()


def load_config(config_file: str) -> Config:
    logger.info(f"Parsing config file {config_file}")
    conf_file = Path(config_file).expanduser().resolve()
    with conf_file.open("r") as f:
        conf_data = yaml.safe_load(f)

    # Validate config file using pydantic schema
    conf = Config(**conf_data)
    return conf


@app.command()
def setup(
    config_file: str = typer.Option(
        ..., "--config", "-c", help="Path to the configuration file."
    ),
    database: Optional[str] = typer.Option(
        None,
        help="Name of the database. Use MetaID from SBML file if not given.",
    ),
    create_db: bool = typer.Option(
        True, help="When database creation is not allowed, set this to False."
    ),
    drop_if_exists: bool = typer.Option(
        False, "--drop-if-exists", "-f", help="Drop the database if it already exists."
    ),
):
    conf = load_config(config_file)

    logger.info("Connecting to neo4j database")
    db = MetaDB(**conf.database.dict())
    meta = Metacyc(db, **conf.metacyc.dict(), db_name=database)

    logger.info("Setting up database using MetaCyc data")
    meta.setup(create_db=create_db, force=drop_if_exists)

    if conf.brenda:
        logger.info("Reading BRENDA text file")
        brenda = parse_brenda(conf.brenda.brenda_file)

    meta.db.close()


@app.command()
def serve():
    """Serves an API to the graph database.

    The command uses the following environment variables to configure the
    database connection: METABOLIKE_DB_URI, METABOLIKE_DB_USER,
    METABOLIKE_DB_PASSWORD, and METABOLIKE_DB_NAME.

    Alternatively, you may save the variables in a file named "metabolike.env"
    in the current directory.
    """
    uvicorn.run(
        "metabolike.api.main:app",
        host="0.0.0.0",
        port=8000,
        log_level="info",
        proxy_headers=True,
    )


if __name__ == "__main__":
    app()



from pydantic import BaseSettings


class DatabaseConfig(BaseSettings):
    metabolike_db_uri: str
    metabolike_db_user: str
    metabolike_db_password: str
    metabolike_db_name: str

    class Config:
        env_file = "metabolike.env"


db_conf = DatabaseConfig()



import logging
from typing import Any, List, Union

from neo4j import BoltDriver, GraphDatabase, Neo4jDriver, Session, Transaction
from neo4j.work.result import Result

logger = logging.getLogger(__name__)


class BaseDB:
    def __init__(
        self,
        uri: str = "neo4j://localhost:7687",
        neo4j_user: str = "neo4j",
        neo4j_password: str = "neo4j",
    ):
        """setup Neo4j driver.

        Args:
            uri: URI of the Neo4j server. Defaults to ``neo4j://localhost:7687``.
            For more details, see :class:`neo4j.driver.Driver`.
            neo4j_user: Neo4j user. Defaults to "neo4j".
            neo4j_password: Neo4j password. Defaults to "neo4j".

        Attributes:
            driver: :class:`neo4j.Neo4jDriver` or :class:`neo4j.BoltDriver`.
            session: :class:`neo4j.Session`.
        """
        self.driver: Union[Neo4jDriver, BoltDriver] = GraphDatabase.driver(
            uri, auth=(neo4j_user, neo4j_password)
        )
        # logger.debug(self.driver.verify_connectivity())

    def close(self):
        self.session.close()
        self.driver.close()

    def create(self, db_name: str, force: bool = False):
        """Helper function to create a database.

        Args:
            db_name: Name of the database to create.
            force: If True, the database will be dropped if it already exists.
        """
        if not db_name:
            raise ValueError("db_name must be specified")
        if force:
            query = f"""CREATE OR REPLACE DATABASE {db_name}"""
        else:
            query = f"""CREATE DATABASE {db_name} IF NOT EXISTS"""
        with self.driver.session() as ss:
            res = ss.run(query).data()

        if res:
            raise RuntimeError(f"Could not create database {db_name}: {res}")

    def start_session(self, **kwargs):
        self.session: Session = self.driver.session(**kwargs)

    def write(self, cypher: str, **kwargs) -> Result:
        """Helper function to write to the database.

        Args:
            query: Query to write to the database.
            **kwargs: Keyword arguments to pass to :meth:`BaseDB.run`.
        """
        return self.session.write_transaction(lambda tx: tx.run(cypher, **kwargs))

    def read(self, cypher: str, **kwargs) -> Result:
        """Helper function to read from the database.

        Args:
            query: Query to read from the database.
            **kwargs: Keyword arguments to pass to :meth:`BaseDB.run`.
        """
        return self.session.read_transaction(self._read_tx_func, cypher, **kwargs)

    @staticmethod
    def _read_tx_func(tx: Transaction, cypher: str, **kwargs) -> List[Any]:
        """Helper function to run a query.

        Args:
            tx: Neo4j transaction function.
            cypher: Cypher query.
            **kwargs: Keyword arguments to pass to the cypher query.

        Returns:
            The result of Cypher query execution.
            For more details, see :class:`neo4j.work.result.Result`.
        """
        res = tx.run(cypher, **kwargs)
        return [x for x in res]



"""Reading and parsing the BRENDA text file.

A lot of the code is translated from the `brendaDb R package`_.
The `BRENDA text file`_ is organised in an EC-number specific format. The
information on each EC-number is given in a very short and compact
way in a part of the file. The contents of each line are described
by a two/three letter acronym at the beginning of the line and
start after a TAB. Empty spaces at the beginning of a line
indicate a continuation line.

.. _`brendaDb R package`: https://bioconductor.org/packages/release/bioc/html/brendaDb.html
.. _`BRENDA text file`: https://www.brenda-enzymes.org/download_brenda_without_registration.php

The contents are organised in ~40 information fields as given
below. Protein information is included in ``#...#``, literature
citations are in ``<...>``, commentaries in ``(...)`` and field-
special information in ``{...}``.

It's not officially documented, but some fields also have commentaries
wrapped in ``|...|``. These are usually for reaction-related fields, so
``(...)`` would be for substrates and ``|...|`` for products.

Protein information is given as the combination organism/Uniprot
accession number where available. When this information is not
given in the original paper only the organism is given.

``///``	indicates the end of an EC-number specific part.
"""

from pathlib import Path
from typing import Dict, List, Optional

import pandas as pd
from lark import Lark
from metabolike.parser.brenda_transformer import *

FIELDS = {
    "ACTIVATING_COMPOUND": "AC",
    "APPLICATION": "AP",
    "CLONED": "CL",
    "COFACTOR": "CF",
    "CRYSTALLIZATION": "CR",
    "ENGINEERING": "EN",
    "EXPRESSION": "EXP",
    "GENERAL_INFORMATION": "GI",
    "GENERAL_STABILITY": "GS",
    "IC50_VALUE": "IC50",
    "INHIBITORS": "IN",
    "KCAT_KM_VALUE": "KKM",  # Kcat/Km value, substrate in {...}
    "KI_VALUE": "KI",  # Ki value, inhibitor in {...}
    "KM_VALUE": "KM",  # Km value, substrate in {...}
    "LOCALIZATION": "LO",
    "METALS_IONS": "ME",
    "MOLECULAR_WEIGHT": "MW",
    "NATURAL_SUBSTRATE_PRODUCT": "NSP",  # reversibility information in {...}
    "ORGANIC_SOLVENT_STABILITY": "OSS",
    "OXIDATION_STABILITY": "OS",
    "PH_OPTIMUM": "PHO",
    "PH_RANGE": "PHR",
    "PH_STABILITY": "PHS",
    "PI_VALUE": "PI",  # isoelectric point
    "POSTTRANSLATIONAL_MODIFICATION": "PM",
    "PROTEIN": "PR",
    "PURIFICATION": "PU",
    "REACTION": "RE",
    "REACTION_TYPE": "RT",
    "RECOMMENDED_NAME": "RN",  # IUPAC accepted name
    "REFERENCE": "RF",
    "RENATURED": "REN",
    "SOURCE_TISSUE": "ST",
    "SPECIFIC_ACTIVITY": "SA",
    "STORAGE_STABILITY": "SS",
    "SUBSTRATE_PRODUCT": "SP",  # reversibility information in {...}
    "SUBUNITS": "SU",
    "SYNONYMS": "SY",
    "SYSTEMATIC_NAME": "SN",
    "TEMPERATURE_OPTIMUM": "TO",
    "TEMPERATURE_RANGE": "TR",
    "TEMPERATURE_STABILITY": "TS",
    "TURNOVER_NUMBER": "TN",  # substrate in {...}
}

# The following fields have additional commentary information,
# reversibility, and a specific reaction format in the descriptions.
FIELD_WITH_REACTIONS = {
    "REACTION",
    "NATURAL_SUBSTRATE_PRODUCT",
    "SUBSTRATE_PRODUCT",
}

# The following fields have substrate or inhibitor information
# wrapped in `{...}`.
FIELD_WITH_SPECIFIC_INFO = {
    "TURNOVER_NUMBER",
    "KCAT_KM_VALUE",
    "KM_VALUE",
    "KI_VALUE",
    "IC50_VALUE",
}

# The following fields don't have descriptions; there is only
# a blob of text wrapped in `(...)`.
FIELD_COMMENTARY_ONLY = {
    "CLONED",
    "CRYSTALLIZATION",
    "PURIFICATION",
    "GENERAL_STABILITY",
    "RENATURED",
    "STORAGE_STABILITY",
}

BASE_GRAMMAR = r"""
%import common.NEWLINE -> _NL

commentary : /\((?=#)/ _separated{content, _COMMENTARY_SEP} /(?<=>)\)/
content    : protein_id description ref_id

!description : TOKENS+
protein_id   : "#" _separated{NUM_ID, _LIST_SEP} "#"
ref_id       : "<" _separated{NUM_ID, _LIST_SEP} ">"

TOKENS: TOKEN+
TOKEN : [_WS] (CHAR+ | INLINE_CHEMICAL) [_WS]
CHAR  : /\w/
      | /[^\x00-\x7F]/  // Unicode characters
      | ARROW | COMPARE_NUM
      | "?" | "." | "," | "+" | ":" | "-" | "/" | "!"  // Common punctuation
      | "*" | "%" | "'" | "\"" | "&" | ";" | "~" | "`" | "\\"  // Rare symbols
      | "[" | "]"  // Wraps chemical names
      | "="  // Appears in REACTION and very rarely in PROTEIN

ARROW          : "<->" | "<-->" | "<-" | "->" | "-->" | "<<" | ">" | /<\t?>/ | /\s?<\s/
               | /<(?=[A-Za-z])/  // Comparison of chemicals in commentary
COMPARE_NUM    : /p\s?<\s?0\.05/i  // p-values

//INLINE_CHEMICAL: /\(+(?!#)/ (CHAR | _WS | COMPARE_NUM | ARROW | "(" | ")")+ ")"
INLINE_CHEMICAL: /\((?!#\d)/ | /(?<!\d>)\)/

_COMMENTARY_SEP : /;\s*(?=#)/
_LIST_SEP       : "," | "\t"
_WS             : ("\t" | " ")+
NUM_ID          : /\d+/
FLOAT_NUM       : NUM_ID "." NUM_ID

_separated{x, sep}: x (sep x)*  // A sequence of 'x sep x sep x ...'

start: entry+
"""


def get_parser_from_field(field: str) -> Optional[Lark]:
    if field == "TRANSFERRED_DELETED":
        return None

    elif field == "REFERENCE":
        grammar = fr"""
            {BASE_GRAMMAR}
            entry      : _ACRONYM ref_id citation [_WS] pubmed [_WS paper_stat] _NL
            citation   : /.+(?={{Pubmed:)/
            pubmed     : /\{{Pubmed:\d*\}}+/
            paper_stat : "(" _separated{{STATS, ","}} ")"
            STATS      : /[\w\/.\-\s]+/
            _ACRONYM: "{FIELDS[field]}\t"
        """
        t = RefTreeTransformer()

    elif field in FIELD_WITH_REACTIONS:
        grammar = fr"""
            {BASE_GRAMMAR}
            entry: _ACRONYM [protein_id] reaction [commentary] [[_WS] more_commentary] [[_WS] reversibility] [[_WS] ref_id] _NL

            reaction: TOKENS+
            %extend CHAR: /\{{(?!(r|ir|\?)?\}})/ | "}}"  // inline curly brackets around chemicals
            %override content: protein_id description [[_WS] reversibility [_WS]] ref_id  // reversibility in commentary

            more_commentary:  "|" _separated{{content, _COMMENTARY_SEP}} "|"
            !reversibility:  "{{}}" | "{{r}}" | "{{ir}}" | "{{?}}"
            _ACRONYM: "{FIELDS[field]}\t"
        """
        t = ReactionTreeTransformer()

    elif field in FIELD_WITH_SPECIFIC_INFO:
        grammar = fr"""
            {BASE_GRAMMAR}
            entry      : _ACRONYM protein_id description substrate [_WS commentary] [_WS] ref_id _NL

            %extend CHAR: /(?<=[^\s])[\{{\}}](?=[^\s])/
            substrate: [_WS] "{{" (TOKENS | "{{" | "}}" | "(" | ")")+ "}}"
            _ACRONYM: "{FIELDS[field]}\t"
        """
        t = SpecificInfoTreeTransformer()

    elif field in FIELD_COMMENTARY_ONLY:
        grammar = fr"""
            {BASE_GRAMMAR}
            entry      : _ACRONYM protein_id _WS [description] ref_id _NL
            %override description: /\(.*\)/ _WS
            _ACRONYM: "{FIELDS[field]}\t"
        """
        t = CommentaryOnlyTreeTransformer()

    else:
        # All other cases can be parsed with the generic grammar.
        # Replace placeholder with actual acronym, and import tokens to be
        # transformed in BaseTransformer.
        grammar = fr"""
            {BASE_GRAMMAR}
            %extend CHAR: "{{" | "}}" // Appears in ENGINEERING and INHIBITORS

            entry      : _ACRONYM [protein_id] description [commentary] [_WS] [ref_id] _NL
            _ACRONYM: "{FIELDS[field]}\t"
        """
        t = GenericTreeTransformer()

    return Lark(grammar, parser="lalr", transformer=t)


def text_to_tree(text: str, parser: Optional[Lark]) -> List[Dict]:
    if parser is None:
        # Simply return the text for TRANSFERRED_DELETED fields
        return [{"description": text}]

    # Parse the text into an annotated tree, then transform into dict
    tree = parser.parse(text)
    return tree.children


def read_brenda(filepath: Path, cache: bool = False) -> pd.DataFrame:
    """Read brenda file and convert to DataFrame.

    Args:
        filepath: Path to the file

    Returns:
        A :class:`pandas.DataFrame` with columns:

           * ID: EC number, e.g. 1.1.1.1
           * field: the content of the information, e.g. protein, localization
           * description: everything else
    """
    cache_file = filepath.with_suffix(".parquet")
    if cache_file.exists() and cache:
        df = pd.read_parquet(cache_file)
        return df

    lines = _read_brenda_file(filepath)
    df = _separate_entries(lines)
    df = _clean_ec_number(df)

    # Fix spefical cases that break the parser
    regexes = [
        ("<Swissprot>", "", False),  # SYNONYMS has this in ref_id
        ("&#8242;", "'", False),  # &#8242; is a single quote
        ("Rmyt2<", "Rmyt2", False),  # Extra < in string
        (
            r"[\x00-\x08\x0b-\x1f\x7f-\x9f]",
            " ",
            True,
        ),  # Remove all control characters other than \t and \n
        # Remove # in text that's not protein ID
        (r"(?<=\w)#(?=\w)", "", True),
        (r"\s#([A-Z])", r" \1", True),
        (r"#(\d+[,\t]?[^\d,\t#])", r"number \1", True),
        (r"([^\s\d\|\(])#", r"\1", True),
        (r"#2#\s#> <38>;", "", True),  # 3.4.21.69
        (
            r"(stereoselectivity) :(.*<5,6,7>) (<3,6>;)",
            r"\1 \3 \2; ",
            True,
        ),  # nested commentary in 1.1.1.270
        (
            r">(4#.+<20>) (<29>)",
            r"\2; #\1",
            True,
        ),  # nested commentary in 2.7.9.1 REACTION
        ("<6> <17>", "<6,17>", False),  # nested commentary in 3.4.21.79 NSP
        ("#2# <4>;", "<4>;", False),  # nested commentary in 5.1.3.3 REACTION
        # Missing opening/closing parenthees
        (r"\sE\)-farnesyl", " (E)-farnesyl", True),
        ("(4a-", "4a-", False),  # (4a-hydroxytetrahydrobiopterin
        ("SN\t2E,6E)", "SN\t(2E,6E)", False),
        (" dU)230 ", " (dU)230 ", False),  # 4.2.99.18 SP
        (
            "carboxyethylsulfanylthiocarbonylamino\t",
            "carboxyethylsulfanylthiocarbonylamino)",
            False,
        ),  # 1.8.1.7 INHIBITORS,
        ("1->4>)", "1->4)", False),  # 4.2.2.8 SUBSTRATE_PRODUCT
        # Replace < with "less than" for non-ref ID cases
        (r"<(\d+[^,\d>])", r"less than \1", True),
        (r"<(\d+,\s\d+)", r"less than \1", True),
        (
            "#1,2,4-8,12-14#",
            "",
            False,
        ),  # Ranges in protein ID for 4.2.1.49
    ]
    for x in regexes:
        df.description = df.description.str.replace(x[0], x[1], regex=x[2])

    df = df[df.description != "SN\n"]  # empty systamatic name

    # Remove redundant `|` in description
    mask = df.description.str.contains(r"\|[^#]") & (
        ~df.field.isin(FIELD_WITH_REACTIONS)
    )
    df.loc[mask, "description"] = df.loc[mask, "description"].str.replace(
        "|", "", regex=False
    )

    # Irregular commentary in 2.2.1.3 REACTION
    mask = (df.ID == "2.2.1.3") & (df.field == "REACTION")
    df.loc[mask, "description"] = df.loc[mask, "description"].str.replace(
        r"\(#.*$", "", regex=True
    )

    df.reset_index(drop=True, inplace=True)

    if cache:
        df.to_parquet(str(cache_file), index=False)

    return df


def parse_brenda(filepath: Union[str, Path], **kwargs) -> pd.DataFrame:

    """Parse the BRENDA text file into a dict.

    This implmentation focuses on extracting information from the text file,
    and feeding the data into a Neo4j database. The parser is implemented
    using Lark. A series of :class:`lark.visitors.Transformer` classes are used
    to clean the data and convert it into the format required by Neo4j.

    Args:
        filepath: The path to the BRENDA text file.
        **kwargs: Additional keyword arguments to pass to :func:`read_brenda`.

    Returns:
        A :class:`pandas.DataFrame` with the ``description`` column from
        :func:`read_brenda` transformed into lists of dicts.
    """
    # Read text file into pandas DataFrame, where the last column contains
    # the text that is to be parsed into trees.
    filepath = Path(filepath).expanduser().resolve()
    df = read_brenda(filepath, **kwargs)

    # Get parsers for each unique field
    parsers: Dict[str, Optional[Lark]] = {"TRANSFERRED_DELETED": None}
    for field in FIELDS.keys():
        parsers[field] = get_parser_from_field(field)

    df["description"] = df.apply(
        lambda row: text_to_tree(row.description, parsers[row.field]),
        axis=1,
    )
    return df

    # TODO: feed the tree into a Neo4j database


def _read_brenda_file(filepath: Path) -> List[str]:
    """Read all non-empty lines from a file.
    Comment lines that start with '*' are ignored. The text file should be downloaded from:
    https://www.brenda-enzymes.org/download_brenda_without_registration.php

    Args:
        filepath: Path to the file

    Returns:
        List of lines in the file.
    """
    if not filepath.is_file():
        raise ValueError(f"Cannot open file: {str(filepath)}")

    res = []
    with open(filepath, "r") as f:
        for line in f:
            # Skip empty lines
            if line == "\n" or line[0] == "*":
                continue

            # Concatenate lines that are split across multiple lines
            if line[0] == "\t":
                res[-1] += f"{line.rstrip()}"
            else:
                res.append(line.strip())

    return res


def _separate_entries(lines: List[str]) -> pd.DataFrame:
    """Convert list of lines to DataFrame.

    Args:
        lines: list of lines from :func:`_read_brenda_file`.

    Returns:
        See :func:`read_brenda`.
    """

    ids, fields, descriptions = [], [], []
    current_id = lines[0][3:]  # remove leading 'ID\t'
    current_field = lines[1]  # a value from FIELD_SET
    ec_info = ""

    # We skip the last line since it's just a separator '///'
    i, n = 2, len(lines) - 1
    while i < n:
        line = lines[i]

        # When we are at the end of an EC-number specific part,
        # we insert the previous entry, update ID and first field, and
        # clear the ec_info.
        if line == "///":
            ids.append(current_id)
            fields.append(current_field)
            descriptions.append(ec_info)
            current_id = lines[i + 1][3:]
            current_field = lines[i + 2]
            ec_info = ""
            i += 2

        # When we are at the end of a field, we insert the previous entry,
        # update field and clear ec_info
        elif line in FIELDS:
            ids.append(current_id)
            fields.append(current_field)
            descriptions.append(ec_info)

            current_field = line
            ec_info = ""

        # # Otherwise we append to the current field
        else:
            ec_info += f"{line}\n"

        i += 1

    # Insert the last entry
    ids.append(current_id)
    fields.append(current_field)
    descriptions.append(ec_info)

    return pd.DataFrame(
        {
            "ID": ids,
            "field": fields,
            "description": descriptions,
        }
    )


def _clean_ec_number(df: pd.DataFrame) -> pd.DataFrame:
    """Remove deleted and transferred EC numbers.

    Some EC numbers have comments wrapped in parentheses. Most of them are
    deleted entries (in this case we remove them) or transferred entries
    (in this case we point to the new entry).

    Args:
        df Generated by :func:`read_brenda`.

    Returns:
        :class:`pandas.DataFrame` with deleted and transferred entries moved to
        the bottom, and their:

           * ``ID`` being the deleted/transferred ID,
           * ``field`` being "TRANSFERRED_DELETED", and
           * ``description`` being the information included in the original ID column.
    """
    df["ID"] = df["ID"].str.replace(" ()", "", regex=False)

    # Only work on entries with parentheses
    df_standard = df[~df["ID"].str.contains("(", regex=False)]
    df_nonstd = df[df["ID"].str.contains("(", regex=False)]
    df_nonstd = (
        df_nonstd.drop_duplicates("ID")
        .assign(
            field="TRANSFERRED_DELETED",
            description=lambda x: x["ID"].str.extract(r"\((.*)\)$", expand=False),
            ID=lambda x: x["ID"].str.extract(r"^((?:\d+\.){3}\d+)", expand=False),
        )
        .query("ID == ID")
    )

    return pd.concat([df_standard, df_nonstd], axis=0, ignore_index=True)
